# SPDX-FileCopyrightText: 2023-present Oak Ridge National Laboratory, managed by UT-Battelle
#
# SPDX-License-Identifier: BSD-3-Clause
import click
import os
import hippogryph as hpg
import enum

from ..__about__ import __version__

class Format(enum.Enum):
    EXO = enum.auto()
    PLOT3D = enum.auto()
    #PLOT3D_3DMB = enum.auto()
    #PLOT3D_3DSB = enum.auto()
    PLOT3D_2D = enum.auto()
    #PLOT3D_2DMB = enum.auto()
    #PLOT3D_2DSB = enum.auto()

def guess_output_format(filename: str) -> Format|None:
    name, extension = os.path.splitext(filename)
    match extension:
        case 'exo':
            return Format.EXO
        case 'xyz':
            return Format.PLOT3D
        case 'xy':
            return Format.PLOT3D_2D
    return None

@click.command()
@click.option('-x', '--x-length', type=click.FloatRange(0.0, min_open=True), show_default=True, default=32.0, help='Length of the grid in the x direction.')
@click.option('-y', '--y-length', type=click.FloatRange(0.0, min_open=True), show_default=True, default=1.0, help='Length of the grid in the y direction.')
#@click.option('-z', '--z-length', type=click.File('w'), show_default=True, default=1.0, help='Length of the grid in the z direction.')
@click.option('-o', '--output', type=click.File('w'), show_default=True, default='hgf.exo', help='Write output to the specified file.')
@click.option('-f', '--format', type=click.Choice(['exo', 'plot3d', 'plot3d_2d']), default=None, help='Specify format to use.')
@click.option('-a', '--ascii', is_flag=True, show_default=True, default=False, help='Write ASCII format (if possible).')
def channel(x_length, y_length, output, format, ascii):
    '''
    Generate a channel grid.
    '''
    block = hpg.Block('domain')
    box = hpg.Box(ni=32, nj=32, block=block, left_label='west', right_label='east',
                  up_label='north', down_label='south')
    mesh = hpg.Mesh('channel')
    mesh.add(box)
    mesh.build()

    xgrid = hpg.Uniform.from_intervals(1.0, x_length)
    ygrid = hpg.Uniform.from_intervals(1.0, y_length, shift=-0.5)
    #xunif = hpg.Uniform.from_delta(ygrid.delta, 16*N)
    #xstretch = hpg.Geometric.from_delta(xunif.delta, 16, N)
    #xgrid = hpg.Composite([xunif, xstretch])

    mesh.apply(xgrid=xgrid, ygrid=ygrid)

def validate_even_int(ctx: click.core.Context,
                  param: click.core.Argument, value: str) -> int:
    try:
        v = int(value)
        if v <= 0:
            raise click.BadParameter('Number of elements must be positive.')
        if v % 2 != 0:
            raise click.BadParameter('Number of elements must be even.')
        return v
    except ValueError:
        raise click.BadParameter('Number of elements must be an integer.')

@click.command()
@click.option('-n', '--number', callback=validate_even_int, show_default=True, default=32, help='Number of elements across the channel (must be even).')
#@click.option('-z', '--z-length', type=click.File('w'), show_default=True, default=1.0, help='Length of the grid in the z direction.')
@click.option('-o', '--output', type=click.Path(writable=True, dir_okay=False), show_default=True, default='hpg.exo', help='Write output to the specified file.')
@click.option('-f', '--format', type=click.Choice(['exo', 'plot3d', 'plot3d_2d']), default=None, help='Specify format to use.')
@click.option('-a', '--ascii', is_flag=True, show_default=True, default=False, help='Write ASCII format (if possible).')
def bfs(number, output, format, ascii):
    '''
    Generate a backward-facing step grid
    '''
    mesh = hpg.backward_step(int(number/2.0))
    mesh.write_exodusii(output)


@click.group(context_settings={'help_option_names': ['-h', '--help']}, invoke_without_command=False)
@click.version_option(version=__version__, prog_name='hippogryph')
@click.pass_context
def hippogryph(ctx: click.Context):
    pass

hippogryph.add_command(channel)
hippogryph.add_command(bfs)


# SPDX-FileCopyrightText: 2014 Jason W. DeGraw <jason.degraw@gmail.com>
# SPDX-FileCopyrightText: 2023-present Oak Ridge National Laboratory, managed by UT-Battelle
#
# SPDX-License-Identifier: BSD-3-Clause
import numpy
#import exodusii

import hippogryph as hpg

def backward_step(M:int):
    N = 2*M
    block = hpg.Block('domain')
    boxN = hpg.Box(ni=17*N, nj=M, block=block, left_label='inflow', right_label='outflow',
                   up_label='north')
    boxS = hpg.Box(ni=17*N, nj=M, block=block, left_label='south', right_label='outflow',
                   down_label='south')
    mesh = hpg.Mesh.from_array('BFS', [boxS, boxN], shape=(1,2))

    mesh.build()

    ygrid = hpg.Uniform.from_intervals(1.0, mesh.nj, shift=-0.75)
    xunif = hpg.Uniform.from_delta(ygrid.delta, 16*N)
    xstretch = hpg.Geometric.from_delta(xunif.delta, 16, N)
    xgrid = hpg.Composite([xunif, xstretch])

    mesh.apply(xgrid=xgrid, ygrid=ygrid)

    return mesh



# SPDX-FileCopyrightText: 2023-present Oak Ridge National Laboratory, managed by UT-Battelle
#
# SPDX-License-Identifier: BSD-3-Clause
from .meshblock import Block, Box, Mesh
from .grid import vkruh, geometric, geometric_sum, single_sided_geometric, single_sided_vinokur, BadGrid, Uniform, Geometric, Composite, VinokurSingleSided
from .cases import backward_step


# SPDX-License-Identifier: 	NASA-1.3

from typing import Dict, List
from .listfunctions import unique_pairs
from .block import Block, reduce_blocks
from .face import Face 
from copy import deepcopy
import math
import numpy as np


def get_outer_faces(block1:Block):
    """Get the outer faces of a block

    Args:
        block1 (Block): A plot3D block

    Returns:
        List[Face]: Non matching faces of the block 
        List[(Face,Face)]: Matching faces inside the block 
    """
    I = [0,block1.IMAX-1]               # Python index starts at 0, need to subtract 1 for it to get the i,j,k
    J = [0,block1.JMAX-1]
    K = [0,block1.KMAX-1]
    # Create the outer faces        
    faces = list()
    face = Face(4)
    i=I[0]
    for j in J:
        for k in K:
            face.add_vertex(block1.X[i,j,k], block1.Y[i,j,k], block1.Z[i,j,k],i,j,k)
    
    faces.append(face)
    face = Face(4)
    i=I[1]
    for j in J:
        for k in K:
            face.add_vertex(block1.X[i,j,k], block1.Y[i,j,k], block1.Z[i,j,k],i,j,k)
    
    faces.append(face)
    face = Face(4)
    j=J[0]
    for i in I:
        for k in K:
            face.add_vertex(block1.X[i,j,k], block1.Y[i,j,k], block1.Z[i,j,k],i,j,k)
    
    faces.append(face)
    face = Face(4)
    j=J[1]
    for i in I:
        for k in K:
            face.add_vertex(block1.X[i,j,k], block1.Y[i,j,k], block1.Z[i,j,k],i,j,k)

    faces.append(face)
    face = Face(4)
    k=K[0]
    for i in I:
        for j in J:
            face.add_vertex(block1.X[i,j,k], block1.Y[i,j,k], block1.Z[i,j,k],i,j,k)
    
    faces.append(face)
    face = Face(4)
    k=K[1]
    for i in I:
        for j in J:
            face.add_vertex(block1.X[i,j,k], block1.Y[i,j,k], block1.Z[i,j,k],i,j,k)
    faces.append(face)

    # Check if faces match each other
    matching = list()
    non_matching = list()
    for i in range(len(faces)):
        matchFound = False
        for j in range(len(faces)):
            if (i!=j and faces[i].vertices_equals(faces[j])):
                matching.append((i,j))
                matchFound = True
        if not matchFound:
            non_matching.append(faces[i]) # these are guaranteed to be exterior 
    matching = list(unique_pairs(matching))
    matching = [(faces[i],faces[j]) for i,j in matching]
    
    # Make sure normals do not intersect 
    # block_center_to_face_center =  block1.cx
    return non_matching, matching # these should be the outer faces
    
def create_face_from_diagonals(block:Block,imin:int,jmin:int,kmin:int,imax:int,jmax:int,kmax:int):
    """Creates a face on a block given a the diagonals defined as (IMIN,JMIN,KMIN), (IMAX, JMAX, KMAX)

    Args:
        block (Block): Block to create a face on 
        imin (int): Lower Corner IMIN
        jmin (int): Lower Corner JMIN
        kmin (int): Lower Corner KMIN
        imax (int): Upper Corner IMAX
        jmax (int): Upper Corner JMAX
        kmax (int): Upper Corner

    Returns:
        (Face): Face created from diagonals 
    """
    newFace = Face(4)           # This is because two of the corners either imin or imax can be equal
    if imin==imax:
        i = imin
        for j in [jmin,jmax]:
            for k in [kmin,kmax]:
                x = block.X[i,j,k]
                y = block.Y[i,j,k]
                z = block.Z[i,j,k]
                newFace.add_vertex(x,y,z,i,j,k)
    elif jmin==jmax:
        j = jmin
        for i in [imin,imax]:
            for k in [kmin,kmax]:
                x = block.X[i,j,k]
                y = block.Y[i,j,k]
                z = block.Z[i,j,k]
                newFace.add_vertex(x,y,z,i,j,k)
    elif kmin==kmax:
        k = kmin
        for i in [imin,imax]:
            for j in [jmin,jmax]:
                x = block.X[i,j,k]
                y = block.Y[i,j,k]
                z = block.Z[i,j,k]
                newFace.add_vertex(x,y,z,i,j,k)
    return newFace

def find_connected_faces(face_to_search:Face,outer_faces:List[Face],connectivity_matrix:np.ndarray,blocks:List[Block]):
    """Recursive program to return all the matching faces. Note faces must have the same I,J,K definition so faces will be matching if for example: Face1 IMIN=IMAX and Face2 IMIN=IMAX and they share a common edge (2 vertices)

    Args:
        face_to_search (Face): This is the face to search for 
        outer_faces (List[Face]): List of outer faces 
        connectivity_matrix (np.ndarray): block connectivity matrix
        searched_faces (List[Face]): Leave this as blank

    Returns:
        List[Face]: list of all faces that connect with face_to_search and it's neighbors. Beware of duplicates.
    """
    connectivity_matrix = connectivity_matrix.copy()
    all_matching_faces = [face_to_search]
    faces_to_search = [face_to_search]
    faces_searched = []
    while len(faces_to_search)>0:
        matching_faces = list()
        for face_to_search in faces_to_search:
            n1 = face_to_search.normal(blocks[face_to_search.BlockIndex])
            selected_block_indx = face_to_search.BlockIndex    
            connected_block_indices = np.where(connectivity_matrix[selected_block_indx,:]==1)[0]
            faces_to_check = [o for o in outer_faces if o.BlockIndex in connected_block_indices.tolist()]
            
            for f in faces_to_check:
                if (len(face_to_search.match_indices(f))==2):
                    n2 = f.normal(blocks[f.BlockIndex])
                    angle = abs(math.degrees(math.acos(np.dot(n1,n2)/(np.linalg.norm(n1)*np.linalg.norm(n2)))))
                    if angle>90:
                        angle = 180-angle

                    if angle<30:
                        connectivity_matrix[selected_block_indx, f.BlockIndex] = 0
                        connectivity_matrix[f.BlockIndex, selected_block_indx] = 0
                        matching_faces.append(f)        
              
        faces_searched.extend(faces_to_search)
        matching_faces = [m for m in matching_faces if m not in all_matching_faces]
        faces_to_search.clear()
        faces_to_search.extend(matching_faces)
        all_matching_faces.extend(matching_faces)        
        matching_faces.clear()
    all_matching_faces = list(set(all_matching_faces))  
    return all_matching_faces


def find_closest_block(blocks:List[Block],x:np.ndarray,y:np.ndarray,z:np.ndarray,centroid:np.ndarray,translational_direction:str="x",minvalue:bool=True):
    """Find the closest block to an extreme in the x,y, or z direction and returns the targetting point. 
    Target point is the reference point where we want the closest block and the closest face 

    Args:
        x (np.ndarray): x coordinate of all the blocks' centroid
        y (np.ndarray): y coordinate of all the blocks' centroid
        z (np.ndarray): z coordinate of all the blocks' centroid
        centroid (np.ndarray): centroid (cx,cy,cz)
        translational_direction (str, optional): _description_. Defaults to "x".
        minvalue (bool, optional): _description_. Defaults to True.

    Returns:
        (tuple): containing

            *selected block index* (int): index of closest block
            *target_x* (float): this is the x value where selected block is closest to
            *target_y* (float): this is the y value where selected block is closest to
            *target_z* (float): this is the z value where selected block is closest to
    """
    cx = centroid[0]; cy = centroid[1]; cz = centroid[2]
    target_x = cx; target_y = cy; target_z = cz
    if translational_direction=="x":
        xmins = [b.X.min() for b in blocks]
        xmaxes = [b.X.max() for b in blocks]
        xmin = min(xmins)
        xmax = max(xmaxes)
        dx = xmax - xmin
        if minvalue:
            target_x = xmin - dx*0.5
            selected_block_indx = np.argmin(np.sqrt((target_x-x)**2 + (cy-y)**2 + (cz-z)**2))            
        else:
            target_x = xmax + dx*0.5
            selected_block_indx = np.argmin(np.sqrt((target_x-x)**2 + (cy-y)**2 + (cz-z)**2))            
        target_y= cy
        target_z = cz
    elif translational_direction=="y":
        ymins = [b.Y.min() for b in blocks]
        ymaxes = [b.Y.max() for b in blocks]
        ymin = min(ymins)
        ymax = max(ymaxes)
        dy = ymax-ymin
        if minvalue:
            target_y = ymin - dy*0.5
            selected_block_indx = np.argmin(np.sqrt((cx-x)**2 + (target_y-y)**2 + (cz-z)**2))
        else:
            target_y = ymax + dy*0.5
            selected_block_indx = np.argmin(np.sqrt((cx-x)**2 + (target_y-y)**2 + (cz-z)**2))
        target_x = cx
        target_z = cz
    else: #  translational_direction=="z":
        zmins = [b.Z.min() for b in blocks]
        zmaxes = [b.Z.max() for b in blocks]
        zmin = min(zmins)
        zmax = max(zmaxes)
        dz = zmax - zmin
        if minvalue:
            target_z = zmin - dz*0.5
            selected_block_indx = np.argmin(np.sqrt((cx-x)**2 + (cy-y)**2 + (target_z-z)**2))            
        else:
            target_z = zmax + dz*0.5
            selected_block_indx = np.argmin(np.sqrt((cx-x)**2 + (cy-y)**2 + (target_z-z)**2))
        target_x = cx
        target_y = cy

    return selected_block_indx,target_x,target_y,target_z 

def find_bounding_faces(blocks:List[Block],connectivity_matrix:np.ndarray,outer_faces:List[Dict[str,int]], direction:str="z"):
    """Finds bounding faces in x,y,z direction so think of this as the xmin and xmax faces of a block.
    This is useful for translational periodicity where you want the bounds to check

    Args:
        blocks (List[Block]): List of blocks
        outer_faces (List[Dict[str,int]]): outer faces as a list of dictionaries 
        direction (str, optional): Direction to search for. Defaults to "z".

    Returns:
        (tuple) containing:
        
            *lower_connected_faces_export* (List[Dict[str,int]]): Export ready version of lower/left connected faces
            *upper_connected_faces_export* (List[Dict[str,int]]): Export ready version of upper/right connected faces
            *lower_connected_faces* (List[face]): List of lower/left connected faces
            *upper_connected_faces* (List[face]): List of upper/right connected faces
    """
    gcd_array = list()
    # Find the gcd of all the blocks
    for block_indx in range(len(blocks)):
        block = blocks[block_indx]
        gcd_array.append(math.gcd(block.IMAX-1, math.gcd(block.JMAX-1, block.KMAX-1)))
    gcd_to_use = min(gcd_array) # You need to use the minimum gcd otherwise 1 block may not exactly match the next block. They all have to be scaled the same way.
    blocks = reduce_blocks(deepcopy(blocks),gcd_to_use)    
    
    xyz_array = np.array([(b.cx, b.cy, b.cz) for b in blocks]); 
    
    cx = np.mean(xyz_array[:,0]) # Centroid 
    cy = np.mean(xyz_array[:,1])
    cz = np.mean(xyz_array[:,2])
    x = xyz_array[:,0]; y = xyz_array[:,1]; z = xyz_array[:,2]

    if len(outer_faces) == 0:
        for i,b in enumerate(blocks):
            outer,_ = get_outer_faces(b)
            [o.set_block_index(i) for o in outer]
            outer_faces.extend(outer)
        outer_faces_all = outer_faces
    else:
        outer_faces_all = outer_face_dict_to_list(blocks,outer_faces,gcd_to_use)

    # Find closest face to the min value 
    selected_block_index,tx,ty,tz = find_closest_block(blocks,x,y,z,np.array([cx,cy,cz]),direction,minvalue=True)
    faces = [f for f in outer_faces_all if f.BlockIndex == selected_block_index]    # Pick all the faces for the selected block 
    min_face_indx = np.argmin(np.array([np.sqrt((f.cx-tx)**2+(f.cy-ty)**2+(f.cz-tz)**2) for f in faces]))
    min_face = faces[min_face_indx]     # Just need this
    
    selected_block_index,tx,ty,tz = find_closest_block(blocks,x,y,z,np.array([cx,cy,cz]),direction,minvalue=False)
    faces = [f for f in outer_faces_all if f.BlockIndex == selected_block_index]    # Pick all the faces for the selected block 
    max_face_indx = np.argmin(np.array([np.sqrt((f.cx-tx)**2+(f.cy-ty)**2+(f.cz-tz)**2) for f in faces]))
    max_face = faces[max_face_indx]     # Also need this

    # Search face connectivity in block connectivity matrix 
    connectivity_matrix = connectivity_matrix - np.eye(connectivity_matrix.shape[0],dtype=np.int8) # turn off connections with itself 

    outer_faces_all = [o for o in outer_faces_all if o.BlockIndex != min_face.BlockIndex]
    outer_faces_all = [o for o in outer_faces_all if o.BlockIndex != max_face.BlockIndex]
    
    # !print("Recursively searching for connected faces")
    lower_connected_faces = find_connected_faces(min_face,outer_faces_all,connectivity_matrix,blocks)
    upper_connected_faces = find_connected_faces(max_face,outer_faces_all,connectivity_matrix,blocks)
    
    lower_connected_faces.append(min_face)
    upper_connected_faces.append(max_face)
    # Unscale faces
    lower_connected_faces = list(set(lower_connected_faces))
    upper_connected_faces = list(set(upper_connected_faces))

    for l in lower_connected_faces:
        l.I*=gcd_to_use
        l.J*=gcd_to_use
        l.K*=gcd_to_use
    for u in upper_connected_faces:
        u.I*=gcd_to_use
        u.J*=gcd_to_use
        u.K*=gcd_to_use

    lower_connected_faces_export = [l.to_dict() for l in lower_connected_faces]
    upper_connected_faces_export = [u.to_dict() for u in upper_connected_faces]
    return lower_connected_faces_export,upper_connected_faces_export, lower_connected_faces, upper_connected_faces

def split_face(face_to_split:Face, block:Block,imin:int,jmin:int,kmin:int,imax:int,jmax:int,kmax:int):
    """Splits a face with another face within the same block 
        picture the split as a two rectangles inside each other

    Args:
        face_to_split (Face): Face on the block to be split 
        block (Block): Block the split is occuring on 
        imin (int): IMIN index of the split (diagonals)
        jmin (int): JMIN index of the split (diagonals)
        kmin (int): KMIN index of the split (diagonals) 
        imax (int): IMAX index of the split
        jmax (int): JMAX index of the split
        kmax (int): KMAX index of the split 

    :: 

                    left face    top face     right face
         ________       __          __            __
        |   __   |     |  |        |__|          |  |     __
        |  |__|  |     |  |         __           |  |    |__|  face_to_split/center face 
        |________|     |__|        |__|          |__|
                                bottom face

    Returns:
        [List[Faces]]: List of unique faces from the split 
    """
    center_face = create_face_from_diagonals(block,
            imin=imin,imax=imax,
            jmin=jmin,jmax=jmax,
            kmin=kmin,kmax=kmax)

    if kmin == kmax:
        # In the picture above Horizontal = i, vertical = j
        left_face = create_face_from_diagonals(block,
                imin=face_to_split.IMIN,imax=imin,
                jmin=face_to_split.JMIN,jmax=face_to_split.JMAX,
                kmin=kmin, kmax=kmax)

        
        right_face = create_face_from_diagonals(block,
                imin=imax, imax=face_to_split.IMAX,
                jmin=face_to_split.JMIN, jmax=face_to_split.JMAX,
                kmin=kmin, kmax=kmax)

        top_face = create_face_from_diagonals(block,
            imin=imin,imax=imax,
            jmin=jmax,jmax=face_to_split.JMAX,
            kmin=kmin,kmax=kmax)
        
        bottom_face = create_face_from_diagonals(block,
            imin=imin,imax=imax,
            jmin=face_to_split.JMIN,jmax=jmin,
            kmin=kmin,kmax=kmax)  

    elif (imin==imax):
        # In the picture above Horizontal = j, vertical = k
        left_face = create_face_from_diagonals(block,
            imin=imin,imax=imax,
            jmin=face_to_split.JMIN, jmax=jmin,
            kmin=face_to_split.KMIN,kmax=face_to_split.KMAX)

        right_face = create_face_from_diagonals(block,
            imin=imin,imax=imax,
            jmin=jmax, jmax=face_to_split.JMAX,
            kmin=face_to_split.KMIN,kmax=face_to_split.KMAX)

        top_face = create_face_from_diagonals(block,
            imin=imin,imax=imax,
            jmin=jmin,jmax=jmax,
            kmin=kmax,kmax=face_to_split.KMAX)

        bottom_face = create_face_from_diagonals(block,
            imin=imin,imax=imax,
            jmin=jmin,jmax=jmax,
            kmin=face_to_split.KMIN,kmax=kmin)

    elif (jmin==jmax):
        # In the picture above Horizontal = i, vertical = k 
        left_face = create_face_from_diagonals(block,
            imin=face_to_split.IMIN,imax=imin,
            jmin=jmin,jmax=jmax,
            kmin=face_to_split.KMIN,kmax=face_to_split.KMAX)

        right_face = create_face_from_diagonals(block,
            imin=imax,imax=face_to_split.IMAX,
            jmin=jmin,jmax=jmax,
            kmin=face_to_split.KMIN,kmax=face_to_split.KMAX)
        
        top_face = create_face_from_diagonals(block,
            imin=imin,imax=imax,
            jmin=jmin,jmax=jmax,
            kmin=kmax,kmax=face_to_split.KMAX)
        
        bottom_face = create_face_from_diagonals(block,
            imin=imin, imax=imax,
            jmin=jmin, jmax=jmax,
            kmin=face_to_split.KMIN, kmax=kmin)
    
    faces = [top_face,bottom_face,left_face,right_face]
    faces = [f for f in faces if not f.isEdge and not f.index_equals(center_face)] # Remove edges
    [f.set_block_index(face_to_split.blockIndex) for f in faces] 
    return faces 

def find_face_nearest_point(faces:List[Face], x:float,y:float,z:float):
    """Find a face nearest to a given point

    Args:
        blocks (List[Block]): List of blocks
        faces (List[Face]): List of faces
        x (float): x coordinate of a reference point
        y (float): y coordinate of a reference point
        z (float): z coordinate of a reference point
    """
    n = list(range(len(faces)))
    dv = list()
    for i in n:
        dx = x-faces[i].cx
        dy = y-faces[i].cy
        dz = z-faces[i].cz
        dv.append(math.sqrt(dx*dx + dy*dy + dz*dz))
    face_index = np.argmin(np.array(dv))
    return face_index

def outer_face_dict_to_list(blocks:List[Block],outer_faces:List[Dict[str,int]],gcd:int=1) -> List[Face]:
    """Converts a list of dictionary face representations to a list of faces. Use this only for outer faces 

    Args:
        blocks (List[Block]): List of blocks
        outer_faces (List[Dict[str,int]]): List of outer faces represented as a dictionary 
        gcd (int, optional): Greatst common divisor. Defaults to 1.

    Returns:
        List[Face]: List of Face objects 
    """
    outer_faces_all = list()
    for o in outer_faces:
        face = create_face_from_diagonals(blocks[o['block_index']], int(o['IMIN']/gcd), int(o['JMIN']/gcd), 
            int(o['KMIN']/gcd), int(o['IMAX']/gcd), int(o['JMAX']/gcd), int(o['KMAX']/gcd))
        if 'id' in o.keys():
            face.id = o['id']
        face.set_block_index(o['block_index'])
        outer_faces_all.append(face)

    return outer_faces_all

def match_faces_dict_to_list(blocks:List[Block],matched_faces:List[Dict[str,int]],gcd:int=1):
    """Converts a list of dictionaries representing matched faces to a list of Faces

    Args:
        blocks (List[Block]): List of blocks 
        matched_faces (List[Dict[str,int]]): List of matched faces represented as a dictionary 
        gcd (int, optional): _description_. Defaults to 1.

    Returns:
        _type_: _description_
    """
    matched_faces_all = list() 
    for _,m in enumerate(matched_faces):
        face1 = create_face_from_diagonals(blocks[m['block1']['block_index']], 
                            int(m['block1']['IMIN']/gcd), int(m['block1']['JMIN']/gcd), int(m['block1']['KMIN']/gcd), 
                            int(m['block1']['IMAX']/gcd), int(m['block1']['JMAX']/gcd), int(m['block1']['KMAX']/gcd))
        face2 = create_face_from_diagonals(blocks[m['block2']['block_index']], 
                            int(m['block2']['IMIN']/gcd), int(m['block2']['JMIN']/gcd), int(m['block2']['KMIN']/gcd), 
                            int(m['block2']['IMAX']/gcd), int(m['block2']['JMAX']/gcd), int(m['block2']['KMAX']/gcd))
        face1.set_block_index(m['block1']['block_index'])
        if 'id' in m['block1'].keys():
            face1.id = m['block1']['id']
        face2.set_block_index(m['block2']['block_index'])
        if 'id' in m['block2'].keys():
            face2.id = m['block2']['id']
        matched_faces_all.append(face1)
        matched_faces_all.append(face2)
    return matched_faces_all


# SPDX-FileCopyrightText: 2023-present Oak Ridge National Laboratory, managed by UT-Battelle
#
# SPDX-License-Identifier: BSD-3-Clause
import argparse
import exodusii
import hippogryph as hpg

def positive_even_int(string):
    value = int(string)
    if value < 0 or value % 2 != 0:
        msg = "%r is not an even, positive integer" % string
        raise argparse.ArgumentTypeError(msg)
    return value

def positive_int(string):
    value = int(string)
    if value < 0:
        msg = "%r is not a positive integer" % string
        raise argparse.ArgumentTypeError(msg)
    return value

def positive_float(string):
    value = float(string)
    if value < 0:
        msg = "%r is not a positive floating point number" % string
        raise argparse.ArgumentTypeError(msg)
    return value

txt = '''Generate T-junction mesh:

             y2 +-------+
                |       |  * H = y1 - y0 = x2 - x1
                |       |  * L = x3 - x2 = y2 - y1
                |       |  * All H length segments are 
                |       |    divided into N equal intervals
                |       |  * d = H/N
                | Block |  * W = x1 - x0 is an integer
                |   4   |    multiple of H
                |       |  * Before/below x2 + L/2 and
                |       |    y1 + L/2, the elements are
                |       |    d x d squares
                |       |  * After/above x2 + L/2 and
                |       |    y1 + L/2, N intervals with
                |       |    d0 = d stretching are used
   y1 +---------+ - - - +-----------------------------------+
      | Block   . Block .           Block                   |
      |   1     .   2   .             3                     |
   y0 +---------+-------+-----------------------------------+
     x0        x1     x2                                  x3   


'''

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description=txt)
    parser.add_argument('-o', '--output', dest='output', action='store',
                        default='tjunct.exo', help='name of output Exodus II files to be write')
    parser.add_argument('-N', '--number', dest='N', action='store',
                        default=32, help='number of intervals across the channel', type=positive_int)
    parser.add_argument('-v', '--verbose', dest='verbose', action='store_true',
                        default=False, help='make lots of noise')
    parser.add_argument('--no-merge', dest='no_merge', action='store_true',
                        default=False, help='keep separate blocks')
    parser.add_argument('-b', metavar='b', dest='length', action='store',
                        default=14, help='sets the branch lengths to L = b*H', type=positive_even_int)
    parser.add_argument('-i', metavar='i', dest='inlet', action='store',
                        default=3, help='sets the inlet length to W = i*H', type=positive_even_int)
    parser.add_argument('-H', dest='H', action='store', type=positive_float,
                        default=1, help='set the height of the inlet, defaults to 1')

    args = parser.parse_args()

    H = args.H
    N = args.N
    half = int(0.5 * args.length)
    Lo2 = half * H
    W = args.inlet * H

    block = hpg.Block('domain')
    inlet = hpg.Box(ni=W*N, nj=N, block=block, left_label='inflow', up_label='inlet_north', down_label='south')
    junction = hpg.Box(ni=N, nj=N, block=block, down_label='south')
    main0 = hpg.Box(ni=half*N, nj=N, block=block, down_label='south', up_label='main_north')
    main1 = hpg.Box(ni=N, nj=N, block=block, down_label='south', right_label='east_outflow', up_label='main_north')
    branch0 = hpg.Box(ni=N, nj=half*N, block=block, left_label='branch_west', right_label='branch_east')
    branch1 = hpg.Box(ni=N, nj=N, block=block, left_label='branch_west', right_label='branch_east', up_label='north_outflow')

    mesh = hpg.Mesh.from_array('T-junction', [inlet, junction, main0, main1, None, branch0, None, None, None, branch1, None, None], shape=(4,3))

    for box in mesh.primitives:
        print(box.i, box.j)

    mesh.build()

    delta = H/N
    xunif = hpg.Uniform.from_delta(delta, inlet.ni + junction.ni + main0.ni)
    xstretch = hpg.Geometric.from_delta(xunif.delta, 7*H, main1.ni)
    xgrid = hpg.Composite([xunif, xstretch])
    yunif = hpg.Uniform.from_delta(delta, inlet.nj + branch0.nj, shift=-0.5*H)
    ystretch = hpg.Geometric.from_delta(yunif.delta, 7*H, branch1.ni)
    ygrid = hpg.Composite([yunif, ystretch])

    mesh.apply(xgrid=xgrid, ygrid=ygrid)

    mesh.save(args.output)

    print(mesh.cell_count)


# SPDX-FileCopyrightText: 2023-present Oak Ridge National Laboratory, managed by UT-Battelle
#
# SPDX-License-Identifier: BSD-3-Clause
import sys

if __name__ == '__main__':
    from .cli import hippogryph

    sys.exit(hippogryph())



# SPDX-License-Identifier: 	NASA-1.3

from .block import Block, reduce_blocks
from .blockfunctions import rotate_block,get_outer_bounds,block_connection_matrix
from .connectivity import find_matching_blocks, get_face_intersection, connectivity_fast, face_matches_to_dict
from .face import Face
from .facefunctions import create_face_from_diagonals, get_outer_faces, find_connected_faces, find_bounding_faces,split_face,find_face_nearest_point,match_faces_dict_to_list,outer_face_dict_to_list,find_closest_block
from .read import read_plot3D, read_ap_nasa
from .write import write_plot3D
from .differencing import find_edges, find_face_edges
from .periodicity import periodicity, periodicity_fast, create_rotation_matrix, rotated_periodicity, translational_periodicity
from .point_match import point_match
from .split_block import split_blocks, Direction
from .listfunctions import unique_pairs



# SPDX-FileCopyrightText: 2023-present Oak Ridge National Laboratory, managed by UT-Battelle
#
# SPDX-License-Identifier: BSD-3-Clause
import argparse
import exodusii
import hippogryph as hpg

def positive_even_int(string):
    value = int(string)
    if value < 0 or value % 2 != 0:
        msg = "%r is not an even, positive integer" % string
        raise argparse.ArgumentTypeError(msg)
    return value

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='Generate backward facing step mesh.')
    parser.add_argument('-o', '--output', dest='output', action='store',
                        default='bfs.exo', help='name of output Exodus II files to be write')
    parser.add_argument('-N', '--number', dest='N', action='store',
                        default=32, help='number of intervals across the channel', type=positive_even_int)
    parser.add_argument('-v', '--verbose', dest='verbose', action='store_true',
                        default=False, help='make lots of noise')

    args = parser.parse_args()

    N = args.N
    M = int(N/2)

    block = hpg.Block('domain')
    boxN = hpg.Box(ni=17*N, nj=M, block=block, left_label='inflow', right_label='outflow',
                   up_label='north')
    boxS = hpg.Box(ni=17*N, nj=M, block=block, left_label='south', right_label='outflow',
                   down_label='south')
    mesh = hpg.Mesh.from_array('BFS', [boxS, boxN], shape=(1,2))

    for box in mesh.primitives:
        print(box.i, box.j)

    mesh.build()

    ygrid = hpg.Uniform.from_intervals(1.0, mesh.nj, shift=-0.75)
    xunif = hpg.Uniform.from_delta(ygrid.delta, 16*N)
    xstretch = hpg.Geometric.from_delta(xunif.delta, 16, N)
    xgrid = hpg.Composite([xunif, xstretch])

    mesh.apply(xgrid=xgrid, ygrid=ygrid)

    mesh.save(args.output)

    print(xgrid.L, xunif.L, xstretch.L)
    print(xgrid.N, xunif.N, xstretch.N)
    #for i in range(xgrid.N+1):
    #    print(i, xgrid.s(i))
    
    print(mesh.cell_count)


# SPDX-License-Identifier: 	NASA-1.3

from copy import deepcopy
from itertools import combinations
import math
import numpy as np
from typing import Dict, List
from tqdm import trange
from .facefunctions import create_face_from_diagonals, get_outer_faces
from .block import Block
from .face import Face

def rotate_block(block,rotation_matrix:np.ndarray) -> Block:
    """Rotates a block by a rotation matrix 

    Args:
        rotation_matrix (np.ndarray): 3x3 rotation matrix 

    Returns:
        Block: returns a new rotated block 
    """
    X = block.X.copy()
    Y = block.Y.copy()
    Z = block.Z.copy()
    points = np.zeros(shape=(3,block.IMAX*block.JMAX*block.KMAX))
    indx = 0
    for i in range(block.IMAX):
        for j in range(block.JMAX):
            for k in range(block.KMAX):
                points[0,indx] = block.X[i,j,k]
                points[1,indx] = block.Y[i,j,k]
                points[2,indx] = block.Z[i,j,k]
                indx+=1
    points_rotated = np.matmul(rotation_matrix,points)
    indx=0
    for i in range(block.IMAX):
        for j in range(block.JMAX):
            for k in range(block.KMAX):
                X[i,j,k] = points_rotated[0,indx]
                Y[i,j,k] = points_rotated[1,indx]
                Z[i,j,k] = points_rotated[2,indx]
                indx+=1
                
    return Block(X,Y,Z)

def reduce_blocks(blocks:List[Block],factor:int):
    """reduce the blocks by a factor of (factor)

    Args:
        blocks (List[Block]): list of blocks to reduce in size
        factor (int, optional): Number of indicies to skip . Defaults to 2.

    Returns:
        [type]: [description]
    """
    for i in range(len(blocks)):
        blocks[i].X = blocks[i].X[::factor,::factor,::factor]
        blocks[i].Y = blocks[i].Y[::factor,::factor,::factor]
        blocks[i].Z = blocks[i].Z[::factor,::factor,::factor]
        blocks[i].IMAX,blocks[i].JMAX,blocks[i].KMAX = blocks[i].X.shape
    return blocks

def get_outer_bounds(blocks:List[Block]):
    """Get outer bounds for a set of blocks

    Args:
        blocks (List[Block]): Blocks defining your shape

    Returns:
        (Tuple) containing: 

            **xbounds** (Tuple[float,float]): xmin,xmax
            **ybounds** (Tuple[float,float]): ymin,ymax
            **zbounds** (Tuple[float,float]): zmin,zmax
    """
    xbounds = [blocks[0].X.min(),blocks[0].X.max()]
    ybounds = [blocks[0].Y.min(),blocks[0].Y.max()]
    zbounds = [blocks[0].Z.min(),blocks[0].Z.max()]
    
    for i in range(1,len(blocks)):
        xmin = blocks[i].X.min()
        xmax = blocks[i].X.max()

        ymin = blocks[i].Y.min()
        ymax = blocks[i].Y.max()

        zmin = blocks[i].Z.min()
        zmax = blocks[i].Z.max()

        if xmin<xbounds[0]:
            xbounds[0] = xmin
        elif xmax>xbounds[1]:
            xbounds[1] = xmax
        
        if ymin<ybounds[0]:
            ybounds[0] = ymin
        elif ymax>ybounds[1]:
            ybounds[1] = ymax

        if zmin<zbounds[0]:
            zbounds[0] = zmin
        elif zmax>zbounds[1]:
            zbounds[1] = zmax
    
    return tuple(xbounds),tuple(ybounds),tuple(zbounds)

def block_connection_matrix(blocks:List[Block],outer_faces:List[Dict[str,int]]=[],tol:float=1E-8):
    """Creates a matrix representing how block edges are connected to each other 

    Args:
        blocks (List[Block]): List of blocks that describe the Plot3D mesh
        outer_faces (List[Dict[str,int]], optional): List of outer faces remaining from connectivity. Useful if you are interested in finding faces that are exterior to the block. Also useful if you combine outerfaces with match faces, this will help identify connections by looking at split faces. Defaults to [].
        tol (float, optional): Matching tolerance to look for when comparing face centroids.
    
    Returns:
        (Tuple): containing

            *connectivity* (np.ndarray): integer matrix defining how the blocks are connected to each other
            *connectivity_i* (np.ndarray): integer matrix defining connectivity of all blocks where IMAX=IMIN
            *connectivity_j* (np.ndarray): integer matrix defining connectivity of all blocks where JMAX=JMIN
            *connectivity_k* (np.ndarray): integer matrix defining connectivity of all blocks where KMAX=KMIN
            
    """
    # Reduce the size of the blocks by the GCD 
    gcd_array = list()    
    for block_indx in range(len(blocks)):
        block = blocks[block_indx]
        gcd_array.append(math.gcd(block.IMAX-1, math.gcd(block.JMAX-1, block.KMAX-1)))
    gcd_to_use = min(gcd_array) # You need to use the minimum gcd otherwise 1 block may not exactly match the next block. They all have to be scaled the same way.
    blocks = reduce_blocks(deepcopy(blocks),gcd_to_use)

    # Face to List 
    outer_faces_all = list()
    for o in outer_faces:
        face = create_face_from_diagonals(blocks[o['block_index']], int(o['IMIN']/gcd_to_use), int(o['JMIN']/gcd_to_use), 
            int(o['KMIN']/gcd_to_use), int(o['IMAX']/gcd_to_use), int(o['JMAX']/gcd_to_use), int(o['KMAX']/gcd_to_use))
        face.set_block_index(o['block_index'])
        if "id" in o:
            face.id = o['id']
        outer_faces_all.append(face)

    outer_faces = outer_faces_all

    n = len(blocks)
    connectivity = np.eye(n,dtype=np.int8)
    combos = list(combinations(range(n),2))    
    for indx in (pbar:=trange(len(combos))):
        i,j = combos[indx]
        pbar.set_description(f"Building block to block connectivity matrix: checking {i}")
        b1 = blocks[i]

        if len(outer_faces)==0:                     # Get the outerfaces to search
            b1_outer_faces,_ = get_outer_faces(b1)
        else:
            b1_outer_faces = [o for o in outer_faces if o.BlockIndex == i]
        
        if i != j and connectivity[i,j]!=-1:
            b2 = blocks[j]

            if len(outer_faces)==0:                 # Get the outerfaces to search
                b2_outer_faces,_ = get_outer_faces(b2)
            else:
                b2_outer_faces = [o for o in outer_faces if o.BlockIndex == j]                

            # Check to see if any of the outer faces of the blocks match   
            connection_found=False             
            for f1 in b1_outer_faces:
                for f2 in b2_outer_faces:
                    if (f1.is_connected(f2,tol)):   # Check if face centroid is the same
                        connectivity[i,j] = 1       # Default block to block connection matrix 
                        connectivity[j,i] = 1
                        connection_found=True
                        # c = np.sum(connectivity[i,:]==1)
                        # print(f"block {i} connections {c}")
                        break
                if connection_found:
                    break
            if not connection_found:
                connectivity[i,j] = -1
                connectivity[j,i] = -1      
    return connectivity



# SPDX-FileCopyrightText: 2014 Jason W. DeGraw <jason.degraw@gmail.com>
# SPDX-FileCopyrightText: 2023-present Oak Ridge National Laboratory, managed by UT-Battelle
#
# SPDX-License-Identifier: BSD-3-Clause
import numpy
import exodusii

class SubBlock:
    def __init__(self, number, i, j, k, ni, nj, nk):
        self.number = number
        self.i = i
        self.j = j
        self.k = k
        self.ni = ni
        self.nj = nj
        self.nk = nk

class Block:
    def __init__(self, name: str, id: int = 1):
        self.name = name
        self.id = id
        self.element_count = 0

class Box:
    def __init__(self, i=0, j=0, k=0, ni=1, nj=1, nk=None, name=None, block=None,
                 front_label=None, back_label=None, left_label=None, right_label=None,
                 up_label=None, down_label=None):
        self.name = name
        self.i = i
        self.j = j
        self.k = k
        self.ni = ni
        self.nj = nj
        self.nk = nk
        if nk is None:
            self.k = 0
            self.nk = 1
            self.two_dimensional = True
        elif k is None:
            self.k = 0
            self.nk = 1
            self.two_dimensional = True
        #self.two_dimensional = k is None or nk is None
        self.block = block
        self.subsets = {}
        self.front = front_label
        if self.front:
            self.add_to_subsets(self.front, 
                                SubBlock(6, self.i, self.j, self.k+self.nk, self.ni, self.nj, 1))
        self.back = back_label
        if self.back:
            self.add_to_subsets(self.back, 
                                SubBlock(5, self.i, self.j, self.k, self.ni, self.nj, 1))
        self.left = left_label
        if self.left:
            self.add_to_subsets(self.left,
                                SubBlock(4, self.i, self.j, self.k, 1, self.nj, self.nk))
        self.right = right_label
        if self.right:
            self.add_to_subsets(self.right, 
                                SubBlock(2, self.i+self.ni, self.j, self.k, 1, self.nj, self.nk))
        self.up = up_label
        if self.up:
            self.add_to_subsets(self.up,
                                SubBlock(3, self.i, self.j + self.nj, self.k, self.ni, 1, self.nk))
        self.down = down_label
        if self.down:
            self.add_to_subsets(self.down,
                                SubBlock(1, self.i, self.j, self.k, self.ni, 1, self.nk))
            
        names = [self.down, self.right, self.up, self.left, self.back, self.front]
        numbers = [1, 2, 3, 4, 5, 6]
        self._sidesets = [(label, number) for label, number in zip(names, numbers) if label is not None]

    def add_to_subsets(self, name, obj):
        if name in self.subsets:
            self.subsets[name].append(obj)
        else:
            self.subsets[name] = [obj]

    def footprint(self):
        return numpy.ones((self.ni, self.nj, self.nk), dtype=numpy.uint8)
    
    def sidesets(self):
        return [el[0] for el in self._sidesets]
    
    def sideset(self, name):
        numbers = [el[1] for el in self._sidesets if el[0] == name]
        results = []
        for number in numbers:
            if number == 1:
                results.append(SubBlock(1, self.i, self.j, self.k, self.ni, 1, self.nk))
            elif number == 2:
                results.append(SubBlock(2, self.i + self.ni - 1, self.j, self.k, 1, self.nj, self.nk))
            elif number == 3:
                results.append(SubBlock(3, self.i, self.j + self.nj - 1, self.k, self.ni, 1, self.nk))
            elif number == 4:
                results.append(SubBlock(4, self.i, self.j, self.k, 1, self.nj, self.nk))
            elif number == 5:
                results.append(SubBlock(5, self.i, self.j, self.k, self.ni, self.nj, 1))
            else:
                results.append(SubBlock(6, self.i, self.j, self.k + self.nk - 1, self.ni, self.nj, 1))
        return results

class AlreadyMeshed(Exception):
    pass

class Mesh:
    def __init__(self, name):
        self.name = name
        self._primitives = []
        self.two_dimensional = False
        self._meshed = False

    @classmethod
    def from_array(cls, name, array, shape=None): # rework this with Numpy or something
        primitives = []
        imax = len(array)
        jmax = 1
        kmax = None
        if shape is not None:
            if len(shape) > 2:
                #print(len(shape))
                raise NotImplementedError("Three dimensional construction from array not implemented")
            imax = shape[0]
            jmax = shape[1]

        # Scan the rows to get the ni and nj values
        i_shift = [None] * imax
        j_shift = [None] * jmax
        index = 0
        for j in range(jmax):
            for i in range(imax):
                if array[index]:
                    if i_shift[i] is None:
                        i_shift[i] = array[index].ni
                    else:
                        if i_shift[i] != array[index].ni:
                            raise NotImplementedError('General dimensioning not implemented, all i sizes in array column must match')
                    if j_shift[j] is None:
                        j_shift[j] = array[index].nj
                    else:
                        if j_shift[j] != array[index].nj:
                            raise NotImplementedError('General dimensioning not implemented, all j sizes in array row must match')
                index += 1
        # Convert those into index shifts
        for i in range(1,imax-1):
            i_shift[i] += i_shift[i-1]
        i_shift.insert(0, 0)
        i_shift = i_shift[:-1]
        for j in range(1,jmax-1):
            j_shift[j] += j_shift[j-1]
        j_shift.insert(0, 0)
        j_shift = j_shift[:-1]
        # Now modify things so everything lines up
        index = 0
        for j in range(jmax):
            row = []
            for i in range(imax):
                row.append(array[index])
                #if array[index]:
                #    jmax = max(jmax, array[index].nj)
                index += 1
            for i,obj in enumerate(row):
                if not obj:
                    continue
                obj.i += i_shift[i]
                obj.j += j_shift[j]
                #if obj.nj != jmax:
                #    raise NotImplementedError
            #last_j += jmax
            primitives.extend([el for el in row if el is not None])
        object = cls(name)
        for primitive in primitives:
            object.add(primitive)

        return object

    @property
    def primitives(self):
        return self._primitives
    
    def add(self, primitive):
        if self._primitives:
            if primitive.two_dimensional == self.two_dimensional:
                self._primitives.append(primitive)
            else:
                return False
        else:
            self._primitives.append(primitive)
            self.two_dimensional = primitive.two_dimensional

    def mesh(self, force=False):
        if self._meshed:
            raise AlreadyMeshed('Mesh "{self.name}" is already meshed')
        if force:
            raise NotImplementedError
        
        self.build(force=force)
    
    def apply(self, xgrid, ygrid, zgrid=None, force=False):
        if self._meshed:
            raise AlreadyMeshed('Mesh "{self.name}" is already meshed')
        if force:
            raise NotImplementedError
        
        if self.two_dimensional:
            x = numpy.zeros(self.ni+1)
            for i in range(self.ni+1):
                x[i] = xgrid.s(i)
                #print('xx', i, x[i])
            y = numpy.zeros(self.nj+1)
            for j in range(self.nj+1):
                y[j] = ygrid.s(j)

            self.x = numpy.zeros(self.node_count)
            self.y = numpy.zeros(self.node_count)

            k = 0
            index = 0
            for j in range(self.nj+1):
                for i in range(self.ni+1):
                    if self.node_index[i, j, k] > 0:
                        self.x[index] = x[i]
                        self.y[index] = y[j]
                        index += 1
        else:
            raise NotImplementedError

    def build(self, force=False):
        if self._meshed:
            raise AlreadyMeshed('Mesh "{self.name}" is already meshed')
        if force:
            raise NotImplementedError
        
        if not self._primitives:
            return
        
        # Get the blocks and sidesets we have
        self.blocks = []
        sidesets = set()
        for primitive in self._primitives:
            if not primitive.block in self.blocks:
                self.blocks.append(primitive.block)
            sidesets.update(primitive.sidesets())
            #print(primitive.sidesets())
        
        # Number the blocks from 1
        reverse_lookup = {}
        for i, block in enumerate(self.blocks):
            block.id = i+1
            reverse_lookup[block.id] = block
        
        # Figure out the size
        if self.two_dimensional:
            i_offset = self._primitives[0].i
            j_offset = self._primitives[0].j
            i_max = self._primitives[0].i + self._primitives[0].ni
            j_max = self._primitives[0].j + self._primitives[0].nj
            for block in self._primitives[1:]:
                i_offset = min(i_offset, block.i)
                j_offset = min(j_offset, block.j)
                i_max = max(i_max, block.i + block.ni)
                j_max = max(j_max, block.j + block.nj)
            self.ni = i_max - i_offset
            self.nj = j_max - j_offset
            self.nk = 1
            self.i_offset = i_offset
            self.j_offset = j_offset
            self.k_offset = 0
        else:
            i_offset = self._primitives[0].i
            j_offset = self._primitives[0].j
            k_offset = self._primitives[0].k
            i_max = self._primitives[0].i + self._primitives[0].ni
            j_max = self._primitives[0].j + self._primitives[0].nj
            k_max = self._primitives[0].k + self._primitives[0].nk
            for block in self._primitives[1:]:
                i_offset = min(i_offset, block.i)
                j_offset = min(j_offset, block.j)
                k_offset = min(k_offset, block.k)
                i_max = max(i_max, block.i + block.ni)
                j_max = max(j_max, block.j + block.nj)
                k_max = max(k_max, block.k + block.nk)
            self.ni = i_max - i_offset
            self.nj = j_max - j_offset
            self.nk = k_max - k_offset
            self.i_offset = i_offset
            self.j_offset = j_offset
            self.k_offset = k_offset

        self.cells = numpy.zeros((self.ni, self.nj, self.nk), dtype=numpy.uint8)
        self.cell_index = numpy.zeros((self.ni, self.nj, self.nk), dtype=numpy.uint64)
        self.node_index = numpy.zeros((self.ni+1, self.nj+1, self.nk+1), dtype=numpy.uint64)

        # Map it out
        for primitive in self.primitives:
            k0 = 0
            k1 = 1
            if not self.two_dimensional:
                k0 = primitive.k - self.k_offset
                k1 = primitive.k + primitive.nk - self.k_offset
            self.cells[primitive.i - self.i_offset:primitive.i + primitive.ni - self.i_offset,
                       primitive.j - self.j_offset:primitive.j + primitive.nj - self.j_offset,
                       k0:k1] = primitive.block.id * primitive.footprint()
        
        # Assign cell numbers and flag nodes, probably need to do this differently
        index = 0
        if self.two_dimensional:
            k = 0
            for j in range(self.nj):
                for i in range(self.ni):
                    if self.cells[i, j, k] > 0:
                        index += 1
                        self.cell_index[i, j, k] = index
                        self.node_index[i,   j,   k] = 1
                        self.node_index[i+1, j,   k] = 1
                        self.node_index[i+1, j+1, k] = 1
                        self.node_index[i,   j+1, k] = 1
                        reverse_lookup[self.cells[i, j, k]].element_count += 1
        else:
            for k in range(self.nk):
                for j in range(self.nj):
                    for i in range(self.ni):
                        if self.cells[i, j, k] > 0:
                            index += 1
                            self.cell_index[i, j, k] = index
                            self.node_index[i,   j,   k] = 1
                            self.node_index[i+1, j,   k] = 1
                            self.node_index[i+1, j+1, k] = 1
                            self.node_index[i,   j+1, k] = 1
                            self.node_index[i,   j,   k+1] = 1
                            self.node_index[i+1, j,   k+1] = 1
                            self.node_index[i+1, j+1, k+1] = 1
                            self.node_index[i,   j+1, k+1] = 1
                            reverse_lookup[self.cells[i, j, k]].element_count += 1
        self.cell_count = index
        # Now number the nodes
        k0 = self.nk + 1
        if self.two_dimensional:
            k0 = 1
        index = 0
        for k in range(k0):
            for j in range(self.nj+1):
                for i in range(self.ni+1):
                    if self.node_index[i, j, k] > 0:
                        index += 1
                        self.node_index[i, j, k] = index
        self.node_count = index

        # Deal with sidesets
        self.sidesets = {}
        for set_name in sidesets:
            self.sidesets[set_name] = []
            for primitive in self._primitives:
                self.sidesets[set_name].extend(primitive.sideset(set_name))
                
        #print(self.cell_count, self.node_count)
        #print(len(self.sidesets))
        #print(self.sidesets)
        #print(self.ni, self.nj)
    
    def write_plot3d(self, filename:str)->bool:
        return False

    def write_exodusii(self, filename:str):
        exo = exodusii.exodusii_file(filename, 'w')
        ndim = 3
        nnodes = 8
        type = 'HEX'
        if self.two_dimensional:
            ndim = 2
            nnodes = 4
            type = 'QUAD'
        exo.put_init(self.name, ndim, self.node_count, self.cell_count,
                     len(self.blocks), 0, len(self.sidesets))
        exo.put_coord(self.x, self.y)

        # Write out the blocks
        for nb, block in enumerate(self.blocks):
            id = nb + 1
            exo.put_element_block(id, type, block.element_count, nnodes)
            exo.put_element_block_name(id, block.name)
            #
            conn = []
            if self.two_dimensional:
                k = 0
                for j in range(self.nj):
                    for i in range(self.ni):
                        if self.cells[i, j, k] == block.id:
                            cell = [self.node_index[i,   j,   k],
                                    self.node_index[i+1, j,   k],
                                    self.node_index[i+1, j+1, k],
                                    self.node_index[i,   j+1, k]]
                            conn.append(cell)
            else:
                for k in range(self.nk):
                    for j in range(self.nj):
                        for i in range(self.ni):
                            if self.cells[i, j, k] > 0:
                                cell = [self.node_index[i,   j,   k],
                                        self.node_index[i+1, j,   k],
                                        self.node_index[i+1, j+1, k],
                                        self.node_index[i,   j+1, k],
                                        self.node_index[i,   j,   k+1],
                                        self.node_index[i+1, j,   k+1],
                                        self.node_index[i+1, j+1, k+1],
                                        self.node_index[i,   j+1, k+1]]
                                conn.append(cell)
            exo.put_element_conn(block.id, numpy.array(conn))

        #exo.close()
        #return

        # Write out the sidesets
        id = 1
        for name, subsets in self.sidesets.items():
            elements = []
            sides = []
            #print(name)
            #print(self.cell_index.shape)
            for sub in subsets:
                #print(range(sub.k - self.k_offset, sub.k - self.k_offset + sub.nk))
                #print(range(sub.j - self.j_offset, sub.j - self.j_offset + sub.nj))
                #print(range(sub.i - self.i_offset, sub.i - self.i_offset + sub.ni))
                for k in range(sub.k - self.k_offset, sub.k - self.k_offset + sub.nk):
                    for j in range(sub.j - self.j_offset, sub.j - self.j_offset + sub.nj):
                        for i in range(sub.i - self.i_offset, sub.i - self.i_offset + sub.ni):
                            if self.cell_index[i,j,k] > 0:
                                elements.append(self.cell_index[i,j,k])
                                sides.append(sub.number)
            #print(len(elements))
            exo.put_side_set_param(id, len(elements))
            exo.put_side_set_name(id, name)
            exo.put_side_set_sides(id, elements, sides)
            id += 1

        exo.close()



# SPDX-FileCopyrightText: 2014 Jason W. DeGraw <jason.degraw@gmail.com>
# SPDX-FileCopyrightText: 2023-present Oak Ridge National Laboratory, managed by UT-Battelle
#
# SPDX-License-Identifier: BSD-3-Clause
import math
import numbers

class BadGrid(Exception):
    pass

def vkruh(factor: float, L: float, i: float, I: int) -> float:
    return L * (1.0 + math.tanh(factor * (i / I - 1.0)) / math.tanh(factor))

def geometric_sum(factor: float, delta: float, i: float) -> float:
    if i == 0:
        return 0.0
    if i == 1:
        return delta
    sum = delta
    for k in range(2, i+1):
        delta *= factor
        sum += delta
    return sum

def geometric(factor: float, delta: float, i: float) -> float:
    if i == 0:
        return 0.0
    if factor == 1:
        return i*delta
    return delta * (1.0 - factor**i)/(1.0 - factor)

def single_sided_geometric(delta: float,
                           I: int, 
                           tolerance: float = 1.0e-14,
                           max_iterations: int = 100,
                           output = print, init=2.0) -> float:
    """
    single_sided_geometric - Given a spacing, determine the required stretching factor

    The geometric one-sided stretching function between 0 and 1 is

            s = sum_k=0,i-1 (delta * factor^k)

    We apply Newton's Method to find the stretching factor given delta and I.The
    stretching function can be reorganized as

            f(factor) = 1.0 + delta - sum_k=1,I-1 (delta * factor^k)

    Then
            f'(factor) = -sum_k=1,I-1 (k * delta * factor^(k-1))
                       = -delta - sum_k=2,I-1 (k * delta * factor^(k-1))
                       = -delta - sum_k=1,I-2 ((k+1) * delta * factor^k)

    There are probably shortcuts, further work is needed. Newton's method can now
    be used to determine the solution.
    """
    output("Geometric Stretching Factor Solution ---------------+")
    output("  ds = % .8e                              |" % delta)
    output("   I = % .4e                                  |" % I)
    output(" tol = % .3e, itermax = %5d                  |" % (tolerance, max_iterations))
    output("----------------------------------------------------+")
    
    # Solve only for positive delta
    if delta <= 0.0:
       output("No solution for negative ds.                        |")
       output("----------------------------------------------------+")
       return None

    factor = init
    if delta > 1.0:
        output("No solution for negative delta > 1.0.               |")
        output("----------------------------------------------------+")
        return None
    #elif I * delta > 1.0:
        #factor /= I * delta
        #f = 1.0
        #factor_power = 1.0
        #for k in range(1,I):
        #    factor_power *= factor
        #    f -= delta * factor_power
    #else:
        #f = 1.0 - (I-1)*delta

    f = delta * (1.0 - factor**I) - 1.0 + factor

    output(" iter          factor                   f           |")
    output("----- ---------------------- ---------------------- |")
    output("%5d % .15e % .15e |" % (1, factor, f))

    for iter in range(2, max_iterations+1):
        #fp = -delta
        #factor_power = 1.0
        #for k in range(1, I-1):
        #    factor_power *= factor
        #    fp -= (k + 1) * delta * factor_power
        #print(fp)

        fp = -delta * I * factor**(I-1) + 1.0

        factor -= f / fp
        #f = 1.0
        #factor_power = 1.0
        #for k in range(1,I):
        #    factor_power *= factor
        #    f -= delta * factor_power

        f = delta * (1.0 - factor**I) - 1.0 + factor

        output("%5d % .15e % .15e |" % (iter, factor, f))

        if abs(f) <= tolerance:
            output("----------------------------------------------------+")
            break
    else:
        output("Failed to converge.                                 |")
        output("----------------------------------------------------+")
        return None
    
    return factor
    

def single_sided_vinokur(ds: float, i: float, I: int, 
                         tolerance: float = 1.0e-14,
                         max_iterations: int = 100,
                         output = print) -> float:
    """
    single_sided_factor - Given a spacing, determine the required stretching factor

    Vinokur's one-sided stretching function between 0 and 1 is

            ds = 1 + tanh(factor * (i / I - 1)) / tanh(factor)

    We apply Newton's Method to find the stretching factor given ds, i, and
    I. The stretching function can be reorganized as

             (ds - 1)tanh(factor) = tanh(factor * (i / I - 1))
                C2 * tanh(factor) = tanh(C1 * factor)

    This only has solutions if |C2| > |C1|.  To see this, note that the
    limit of the LHS is C2, and the limit of the RHS is -1.  Both C1 and C2
    are negative.  The slopes at zero are C2 and C1. respectively, so the
    only possible way that the two curves can intersect is if |C2| > |C1|.
    Approximating tanh(C1*delta) as the line C1*delta for small delta, and
    approximating  C2*tanh(delta) as C2 for large delta, we obtain an first
    guess for the answer as

                               C2 = C1*factor

    Newton's method can now be used to determine the solution.
    """

    C1 = float(i) / float(I) - 1.0
    C2 = ds - 1.0
    
    output("Vinokur h Stretching Factor Solution ---------------+")
    output("  ds = % .8e                              |" % ds)
    output("   I = % .4e                                  |" % I)
    output(" tol = % .3e, itermax = %5d                  |" % (tolerance, max_iterations))
    output("----------------------------------------------------+")
    
    # Solve only for positive ds
    if ds <= 0.0:
       output("No solution for negative ds.                        |")
       output("----------------------------------------------------+")
       return None

    # Only have a nontrivial solution for |C2| > |C1|
    if (C2 > 0.0) or (C2 > C1):
        output("No solution for given inputs.                       |")
        output("----------------------------------------------------+")
        return None
    
    # As an initial guess, take the intersection of the lines
    #      y = 1.0   (limit of tanh(d*C1)
    #      y = C2*d  (approximation of C2*tanh(d))
    #      y = C2    (limit of C2*tanh(d))
    #      y = C1*d  (approximation of tanh(C1*d))

    factor = C1 / C2
    f = math.tanh(factor * C1) - C2 * math.tanh(factor)
    output(" iter          factor                   f           |")
    output("----- ---------------------- ---------------------- |")
    output("%5d % .15e % .15e |" % (1, factor, f))

    for iter in range(2, max_iterations+1):
        s1 = 1.0 / math.cosh(factor * C1)
        s2 = 1.0 / math.cosh(factor)
        fp = C1 * s1 * s1 - C2 * s2 * s2
        factor -= f / fp
        f = math.tanh(factor * C1) - C2 * math.tanh(factor)

        output("%5d % .15e % .15e |" % (iter, factor, f))

        if abs(f) <= tolerance:
            output("----------------------------------------------------+")
            break
    else:
        output("Failed to converge.                                 |")
        output("----------------------------------------------------+")
        return None
    
    return factor

class Uniform:
    def __init__(self, delta: float, L: float, N: int, shift: float = 0.0):
        self.delta = delta
        self.L = L
        self.N = N
        self.shift = shift

    def s(self, i:float) -> float:
        return self.shift + i * self.delta
    
    @classmethod
    def from_json(cls, **kwargs):
        shift = kwargs.get('shift', 0.0)
        N = kwargs.get('N', None)
        try:
            N = int(N)
        except ValueError:
            return None
        delta = kwargs.get('delta', None)
        if isinstance(delta, numbers.Number):
            L = kwargs.get('L', None)
            if isinstance(L, numbers.Number):
                return cls(delta, L, N, shift=shift)
            else:
                return cls.from_delta(delta, N, shift=shift)
        else:
            L = kwargs.get('L', None)
            if isinstance(L, numbers.Number):
                return cls.from_intervals(L, N, shift=shift)
        return None
    
    @classmethod
    def from_delta(cls, delta: float, N: int, shift: float = 0.0):
        L = delta * N
        return cls(delta, L, N, shift=shift)
    
    @classmethod
    def from_intervals(cls, L: float, N: int, shift: float = 0.0):
        delta = L / float(N)
        return cls(delta, L, N, shift=shift)

class VinokurSingleSided:
    def __init__(self, factor: float, L: float, N: int, shift: float = 0.0):
        self.factor = factor
        self.N = N
        self.I = N # The number of intervals is also where the function ends
        self.L = L
        self.shift = shift

    def s(self, i: float) -> float:
        return self.shift + vkruh(self.factor, self.L, i, self.I)
    
    @classmethod
    def from_json(cls, **kwargs):
        shift = kwargs.get('shift', 0.0)
        factor = kwargs.get('factor', None)
        if factor is None: # Should really check that it's a number
            if all(name in kwargs for name in ('delta', 'i', 'N', 'L')):
                tolerance = kwargs.get('tolerance', 1.0e-14)
                max_iterations = kwargs.get('max_iterations', 100)
                return cls.from_delta(kwargs['delta'], kwargs['L'], kwargs['i'], kwargs['N'],
                                       tolerance=tolerance, max_iterations=max_iterations, shift=shift)
        elif all(name in kwargs for name in ('N', 'L')):
            return cls(factor, kwargs['L'], kwargs['N'], shift=shift)
        return None
    
    @classmethod
    def from_delta(cls, delta: float, L: float, i: float, I: int, tolerance: float = 1.0e-14, max_iterations: int = 100,
                   output=print, shift: float = 0.0):
        ds = delta / L # Rescale
        factor = single_sided_vinokur(ds, i, I, tolerance=tolerance, max_iterations=max_iterations, output=output)
        if factor is None:
            return None
        return cls(factor, L, I, shift=shift)
    
class Geometric:
    def __init__(self, factor: float, delta: float, L: float, N: int, shift: float = 0.0):
        self.factor = factor
        self.delta = delta
        self.L = L
        self.N = N
        self.shift = shift

    def s(self, i: float) -> float:
        return self.shift + geometric(self.factor, self.delta, i)
    
    @classmethod
    def from_json(cls, **kwargs):
        shift = kwargs.get('shift', 0.0)
        if all(name in kwargs for name in ('delta', 'N')):
            if all(name in kwargs for name in ('factor', 'L')):
                return cls(kwargs['factor'], kwargs['delta'], kwargs['L'], kwargs['N'], shift=shift)
            L = kwargs.get('L', 1.0)
            tolerance = kwargs.get('tolerance', 1.0e-14)
            max_iterations = kwargs.get('max_iterations', 100)
            return cls.from_delta(kwargs['delta'], L, kwargs['N'], tolerance=tolerance, max_iterations=max_iterations,
                                  shift=shift)
    
    @classmethod
    def from_delta(cls, delta: float, L: float, I: int, tolerance: float = 1.0e-14, max_iterations: int = 100,
                   output=print, shift: float = 0.0):
        ds = delta / L # Rescale
        factor = single_sided_geometric(ds, I, tolerance=tolerance, max_iterations=max_iterations, output=output)
        if factor is None:
            return None
        return cls(factor, delta, L, I, shift=shift)
    
class Composite:
    def __init__(self, grids):
        self.grids = grids
        if not grids:
            raise BadGrid('Composite grid cannot be constructed without input grids.')
        self.L = 0.0
        self.N = 0
        self.intervals = []
        self.shift = grids[0].shift
        for grid in self.grids:
            self.N += grid.N
            grid.shift = self.L
            self.L += grid.L
            self.intervals.append(self.N)

    def s(self, i: float) -> float:
        last = 0
        for k, n in enumerate(self.intervals[:-1]):
            if i <= n:
                return self.shift + self.grids[k].s(i-last)
            last = n
        if i == self.N:
            return self.shift + self.L
        return self.shift + self.grids[-1].s(i-last)


# SPDX-License-Identifier: 	NASA-1.3


def unique_pairs(listOfItems=[]):
    """Checks if an item is not already in the list 
    
    Args:
        listOfItems (list): list of combinations e.g. (1,2),(3,4),(2,1)

    Yields:
        unique pair: [description]
    """
    seen = set()  #use set to keep track of already seen items, sets provide O(1) lookup  
    for x,y in listOfItems:
        if x!=y and (y,x) not in seen:
            seen.add((x,y)) 
            yield x,y



# SPDX-FileCopyrightText: 2014 Jason W. DeGraw <jason.degraw@gmail.com>
# SPDX-FileCopyrightText: 2023-present Oak Ridge National Laboratory, managed by UT-Battelle
#
# SPDX-License-Identifier: BSD-3-Clause
import hippogryph
import pytest

def dev_null(mesg):
    pass

def test_one_sided_vinokur_functions():
    ds = hippogryph.single_sided_vinokur(1.0e-6, 1, 16, output=dev_null)
    assert abs(ds - 7.4717023123662765) < 1.0e-15
    assert abs(hippogryph.vkruh(ds, 1.0, 1, 16) - 1.0e-6) < 1.0e-15

    ds = hippogryph.single_sided_vinokur(-1.0e-6, 1, 16, output=dev_null)
    assert ds is None

    ds = hippogryph.single_sided_vinokur(2.0, 1, 16, output=dev_null)
    assert ds is None

    ds = hippogryph.single_sided_vinokur(1.0e-6, 1, 16, output=dev_null, max_iterations=2)
    assert ds is None

def test_one_sided_vinokur_object():
    vkr = hippogryph.VinokurSingleSided.from_delta(1.0e-6, 1.0, 1, 16, output=dev_null)
    assert abs(vkr.s(1) - 1.0e-6) < 1.0e-15
    assert abs(vkr.s(16) - 1.0) < 1.0e-15

    vkr = hippogryph.VinokurSingleSided.from_delta(2.0, 1.0, 1, 16, output=dev_null)
    assert vkr is None

def test_uniform_object():
    obj = hippogryph.Uniform.from_delta(0.1, 10)
    assert obj.L == 1.0
    assert obj.s(1) == obj.delta
    obj = hippogryph.Uniform.from_intervals(1.0, 10)
    assert obj.delta == 0.1
    assert obj.s(1) == obj.delta

def test_geometric_function():
    factor = hippogryph.single_sided_geometric(1.0e-3, 32, output=dev_null)
    assert abs(hippogryph.geometric(factor, 1.0e-3, 32) - 1.0) < 1.0e-14
    assert abs(hippogryph.geometric_sum(factor, 1.0e-3, 32) - 1.0) < 1.0e-14
    assert hippogryph.geometric(factor, 1.0e-3, 0) == 0.0
    assert hippogryph.geometric(factor, 1.0e-3, 1) == 1.0e-3
    assert hippogryph.geometric(1.0, 0.1, 2) == 0.2
    assert hippogryph.geometric_sum(factor, 1.0e-3, 0) == 0.0
    assert hippogryph.geometric_sum(factor, 1.0e-3, 1) == 1.0e-3

def test_geometric_object():
    obj = hippogryph.Geometric.from_delta(0.025, 10.0, 32, output=dev_null)
    #assert abs(obj.factor - 11.328501284496937) < 1.0e-14
    assert obj.s(1) == 0.025
    assert obj.s(2) == 0.025 * (1.0 + obj.factor)
    assert abs(obj.s(32) - 10.0) < 1.0e-14

def test_composites():
    # Two uniform grids
    grids = [hippogryph.Uniform.from_delta(0.1, 10),
             hippogryph.Uniform.from_delta(0.2, 5)]
    obj = hippogryph.Composite(grids)
    assert obj.L == 2.0
    assert obj.N == 15
    assert obj.s(10) == 1.0
    assert obj.s(11) == 1.2

    grids[1] = hippogryph.Geometric.from_delta(0.025, 10.0, 32, output=dev_null)
    obj = hippogryph.Composite(grids)
    assert obj.L == 11.0
    assert obj.N == 42
    assert obj.s(42) == 11.0
    assert obj.intervals == [10, 42]

    grids = []
    with pytest.raises(hippogryph.BadGrid):
        obj = hippogryph.Composite(grids)

def test_fails():
    factor = hippogryph.single_sided_geometric(-0.002, 32, output=dev_null)
    assert factor is None
    factor = hippogryph.single_sided_geometric(2.0, 32, output=dev_null)
    assert factor is None
    grid = hippogryph.Geometric.from_delta(25.0, 10.0, 0, 32, output=dev_null)
    assert grid is None
    factor = hippogryph.single_sided_geometric(1.0e-3, 32, output=dev_null, max_iterations=2)
    assert factor is None


# SPDX-License-Identifier: 	NASA-1.3

import numpy as np 
import os.path as osp
import struct
from typing import List
from .block import Block
from scipy.io import FortranFile
from tqdm import trange

def __read_plot3D_chunk_binary(f,IMAX:int,JMAX:int,KMAX:int, big_endian:bool=False,read_double:bool=True):
    """Reads and formats a binary chunk of data into a plot3D block

    Args:
        f (io): file handle
        IMAX (int): maximum I index
        JMAX (int): maximum J index
        KMAX (int): maximum K index
        big_endian (bool, Optional): Use big endian format for reading binary files. Defaults False.

    Returns:
        numpy.ndarray: Plot3D variable either X,Y, or Z 
    """
    A = np.empty(shape=(IMAX, JMAX, KMAX))
    for k in range(KMAX):
        for j in range(JMAX):
            for i in range(IMAX):
                if read_double:
                    A[i,j,k] = struct.unpack(">d",f.read(8))[0] if big_endian else struct.unpack("<d",f.read(8))[0]
                else:
                    A[i,j,k] = struct.unpack(">f",f.read(4))[0] if big_endian else struct.unpack("<f",f.read(4))[0]
    return A

def read_word(f):
    """Continously read a word from an ascii file

    Args:
        f (io): file handle

    Yields:
        float: value from ascii file
    """
    for line in f:
        line = line.strip().replace('\n','').split(' ')
        tokenArray = [float(entry) for entry in line if entry]
        for token in tokenArray:
            yield token

def __read_plot3D_chunk_ASCII(f,IMAX:int,JMAX:int,KMAX:int):
    """Reads and formats a binary chunk of data into a plot3D block

    Args:
        f (io): file handle
        IMAX (int): maximum I index
        JMAX (int): maximum J index
        KMAX (int): maximum K index
        big_endian (bool, Optional): Use big endian format for reading binary files. Defaults False.

    Returns:
        numpy.ndarray: Plot3D variable either X,Y, or Z 
    """
    tokenArray = np.zeros(shape=(IMAX*JMAX*KMAX))
    i = 0
    for w in read_word(f):
        tokenArray[i] = w
        i+=1
        if i>len(tokenArray)-1:
            break

    A = np.reshape(tokenArray,newshape=(KMAX,JMAX,IMAX))
    A = np.transpose(A,[2,1,0])    
    return A

def read_ap_nasa(filename:str):
    """Reads an AP NASA File and converts it to Block format which can be exported to a plot3d file
        AP NASA file represents a single block. The first 7 integers are il,jl,kl,ile,ite,jtip,nbld
    
    Args:
        filename (str): location of the .ap file

    Returns:
        Tuple containing: 

            *block* (Block): file in block format
            *nbld* (int): Number of blades 
    """

    f = FortranFile(filename, 'r')

    ints = f.read_ints(np.int32)
    idim = np.array([ints[0],ints[1],ints[2]])
    mdim = np.array([3,ints[0]*ints[2]])
    il = ints[0]
    jl = ints[1]
    kl = ints[2]
    jdim = jl 

    ile = ints[3]
    ite = ints[4]
    jtip = ints[5]
    nbld = ints[6]

    for j in range(0,jdim):
        jmeshxrt = f.read_reals(dtype='f4').reshape(mdim)
        meshi    = np.array(jmeshxrt[0,:])
        meshj    = np.array(jmeshxrt[1,:])
        meshk    = np.array(jmeshxrt[2,:])
        if j == 0:
            meshx   = meshi
            meshr   = meshj
            mesht   = meshk
        else:
            meshx = np.append(meshx,meshi)
            meshr = np.append(meshr,meshj)
            mesht = np.append(mesht,meshk)

    meshx = meshx.reshape(ints[1],ints[2],ints[0])
    meshr = meshr.reshape(ints[1],ints[2],ints[0])
    mesht = mesht.reshape(ints[1],ints[2],ints[0])

    # Convert from x,r,theta to x,y,z
    z = meshr*np.sin(mesht)
    y = meshr*np.cos(mesht)

    return Block(X=meshx,Y=y,Z=z), nbld


def read_plot3D(filename:str, binary:bool=True,big_endian:bool=False,read_double:bool=True):
    """Reads a plot3d file and returns Blocks

    Args:
        filename (str): name of the file to read, .p3d, .xyz, .pdc, .plot3d? 
        binary (bool, optional): indicates if the file is binary. Defaults to True.
        big_endian (bool, optional): use big endian format for reading binary files
        read_float (bool, optional): read floating point. Only affects binary files

    Returns:
        List[Block]: List of blocks insdie the plot3d file
    """
    
    blocks = list()
    if osp.isfile(filename):
        if binary:
            with open(filename,'rb') as f:
                nblocks = struct.unpack(">I",f.read(4))[0] if big_endian else struct.unpack("I",f.read(4))[0] # Read bytes            
                IMAX = list(); JMAX = list(); KMAX = list()
                for b in range(nblocks):
                    if big_endian:
                        IMAX.append(struct.unpack(">I",f.read(4))[0]) # Read bytes
                        JMAX.append(struct.unpack(">I",f.read(4))[0]) # Read bytes
                        KMAX.append(struct.unpack(">I",f.read(4))[0]) # Read bytes
                    else:
                        IMAX.append(struct.unpack("I",f.read(4))[0]) # Read bytes
                        JMAX.append(struct.unpack("I",f.read(4))[0]) # Read bytes
                        KMAX.append(struct.unpack("I",f.read(4))[0]) # Read bytes

                for b in range(nblocks):
                    X = __read_plot3D_chunk_binary(f,IMAX[b],JMAX[b],KMAX[b], big_endian,read_double)
                    Y = __read_plot3D_chunk_binary(f,IMAX[b],JMAX[b],KMAX[b], big_endian,read_double)
                    Z = __read_plot3D_chunk_binary(f,IMAX[b],JMAX[b],KMAX[b], big_endian,read_double)
                    b_temp = Block(X,Y,Z)                    
                    blocks.append(b_temp)
        else:
            with open(filename,'r') as f: 
                nblocks = int(f.readline())
                IMAX = list(); JMAX = list(); KMAX = list()
                
                for b in range(nblocks):
                    IJK = f.readline().replace('\n','').split(' ')
                    tokens = [int(w) for w in IJK if w]
                    IMAX.append(tokens[0])
                    JMAX.append(tokens[1])
                    KMAX.append(tokens[2])            

                for b in trange(nblocks):
                    X = __read_plot3D_chunk_ASCII(f,IMAX[b],JMAX[b],KMAX[b])
                    Y = __read_plot3D_chunk_ASCII(f,IMAX[b],JMAX[b],KMAX[b])
                    Z = __read_plot3D_chunk_ASCII(f,IMAX[b],JMAX[b],KMAX[b])
                    b_temp = Block(X,Y,Z)                    
                    blocks.append(b_temp)
    return blocks


